#coding=utf-8
import os.path as osp
import open3d as o3d
import argparse
import torch
import torch.nn.functional as F
from torch_geometric.datasets import ShapeNet
from torch_geometric.data import Data
import torch_geometric.transforms as T
#from rgcn_conv import RGCNConv
from torch_geometric.nn import GCNConv, DynamicEdgeConv,knn_graph, EdgeConv, radius_graph, RGCNConv# noqa
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import random
#import h5py
import numpy as np
from collections import Counter
from tqdm import tqdm
from EigenGraph import eigen_Graph
from data import ShapeNetPart
from scipy.stats import mode
random.seed(100)
torch.manual_seed(7)
np.random.seed(100)
from torch.nn import Sequential as Seq, Dropout, Linear as Lin,BatchNorm1d as BN
def MLP(channels, bias=True):
    return Seq(*[
        Seq(Lin(channels[i - 1], channels[i], bias=bias),BN(channels[i]),torch.nn.LeakyReLU(negative_slope=0.2),Lin(channels[i], channels[i], bias=bias))
        for i in range(1, len(channels))
    ])

class RelationGraphConvNet(torch.nn.Module):
    def __init__(self,num_points, k = 20,num_bases = 5):
        super(RelationGraphConvNet, self).__init__()
        self.conv1 = RGCNConv(3, 16, num_relations = 2, num_bases = num_bases)
        #self.conv2 = RGCNConv(16, 8, num_relations = 3, num_bases = num_bases)
        self.conv2 = RGCNConv(16, num_classes, num_relations = 2, num_bases = num_bases)
        
        self.k = k
        self.num_points = num_points

    def forward(self):
        x = data.x
        ##############r-gcn
        x = F.relu(self.conv1(x, edge_index, edge_weight))
        x = F.dropout(x, training=self.training)
        #x = F.relu(self.conv2(x, edge_index, edge_weight))
        #x = F.dropout(x, training=self.training)
        x = self.conv2(x, edge_index, edge_weight)

        return F.log_softmax(x, dim=1)


class GCN(torch.nn.Module):
    def __init__(self):
        super(GCN, self).__init__()
        self.conv1 = GCNConv(point_features, 16, cached=True)
        self.conv2 = GCNConv(16, num_classes, cached=True)
        # self.conv1 = ChebConv(data.num_features, 16, K=2)
        # self.conv2 = ChebConv(16, data.num_features, K=2)

    def forward(self):
        x= data.x
        x = F.relu(self.conv1(x, edge_index))
        x = F.dropout(x, training=self.training)
        x = self.conv2(x, edge_index)
        return F.log_softmax(x, dim=1)


import os
import sys
import glob
import h5py
import numpy as np
from torch.utils.data import Dataset


def center_data(pcs):
	for pc in pcs:
		centroid = np.mean(pc, axis=0)
		pc[:,0]-=centroid[0]
		pc[:,1]-=centroid[1]
		pc[:,2]-=centroid[2]
	return pcs

def normalize_data(pcs):
	for pc1 in pcs:
		#get furthest point distance then normalize
		d = max(np.sum(np.abs(pc1)**2,axis=-1)**(1./2))
		pc1 /= d

		# pc[:,0]/=max(abs(pc[:,0]))
		# pc[:,1]/=max(abs(pc[:,1]))
		# pc[:,2]/=max(abs(pc[:,2]))

	return pcs


def load_data(partition):

    with h5py.File(".//data//realworlddataset//main_split_hardest//"+'%s_objectdataset_augmentedrot_scale75.h5' % partition,'r')  as f:
    #for h5_name in glob.glob(os.path.join('..', 'data//realworlddataset//main_split_nobg', '%s*.h5' % partition)):
        #f = h5py.File(h5_name)
        data = f['data'][:].astype('float32')
        label = f['label'][:].astype('int64')
        mask = f['mask'][:].astype('int64')
        f.close()
    #draw_Point_Cloud(data[111])
    data = center_data(data)
    data = normalize_data(data)
    #draw_Point_Cloud(data[111])
    print("load real data and center and normalize data")
    print(data.shape, label.shape, mask.shape)
    return data, label,mask


def translate_pointcloud(pointcloud):
    xyz1 = np.random.uniform(low=2. / 3., high=3. / 2., size=[3])
    xyz2 = np.random.uniform(low=-0.2, high=0.2, size=[3])

    translated_pointcloud = np.add(np.multiply(pointcloud, xyz1), xyz2).astype('float32')
    return translated_pointcloud


def jitter_pointcloud(pointcloud, sigma=0.01, clip=0.02):
    N, C = pointcloud.shape
    pointcloud += np.clip(sigma * np.random.randn(N, C), -1 * clip, clip)
    return pointcloud


class RealPointCloudRaw(Dataset):
    def __init__(self, num_points, partition='test', shuffle = True):
        self.data, self.label,self.mask= load_data(partition)
        self.num_points = num_points
        self.partition = partition
        self.shuffle = shuffle
    def __getitem__(self, item):
        # raw data num_points is 2048
        # idx = np.arange(2048)
        # np.random.shuffle(idx)
        # #sample to num_points
        # pointcloud = self.data[item][idx[:self.num_points]]
        label = self.label[item]
        mask = self.mask[item][:self.num_points]
        pointcloud = self.data[item][:self.num_points]
        if self.shuffle and self.partition == 'train':
            indices = list(range(pointcloud.shape[0]))
            np.random.shuffle(indices)
            pointcloud = pointcloud[indices]
            mask = mask[indices]
        return pointcloud, label,mask

    def __len__(self):
        return self.data.shape[0]


def keypoint_detect(point_cloud, mask, sample_ratio=0.01):
    '''
    Compute ISS Keypoints on Armadillo
    :param mask: used to guarantee keypoints are from both foreground and background
    :return:
    '''
    pcd = o3d.geometry.PointCloud()
    pcd.points = o3d.utility.Vector3dVector(point_cloud)

    # MIN_NEIGHBORS = 5
    # SALIENT_RADIUS = 0.0
    # NON_MAX_RADIUS = 0.0
    # GAMMA_21 = 0.5
    # GAMMA_32 = 0.75

    MIN_NEIGHBORS = 5
    SALIENT_RADIUS = 0.0
    NON_MAX_RADIUS = 0.0
    GAMMA_21 = 0.975
    GAMMA_32 = 0.975

    keypoint = o3d.geometry.keypoint.compute_iss_keypoints(input=pcd, salient_radius=SALIENT_RADIUS, non_max_radius=NON_MAX_RADIUS, gamma_21=GAMMA_21, gamma_32=GAMMA_32, min_neighbors=MIN_NEIGHBORS)
    keypoint = np.asarray(keypoint.points)
    full_keypoint_idx = np.array([np.argwhere(point_cloud == keypoint[i])[0, 0] for i in range(len(keypoint))])

    if len(full_keypoint_idx) > int(sample_ratio * len(point_cloud)):
        sample_from_keypoint = int(sample_ratio * len(point_cloud))
    else:
        sample_from_keypoint = len(full_keypoint_idx)

    if sample_from_keypoint == 0:
        print("Inappropriate parameters was set!\n")
        full_keypoint_idx = np.random.choice(len(point_cloud), 1)
        sample_from_keypoint = 1

    if (0 not in mask) or (1 not in mask):
        keypoint_idx = np.random.choice(full_keypoint_idx, sample_from_keypoint, replace=False)
    else:
        if (0 not in mask[full_keypoint_idx]) or (1 not in mask[full_keypoint_idx]):
            if 0 not in mask[full_keypoint_idx]:
                addition_idx = np.random.choice(np.argwhere(mask == 0).reshape(-1,), 1)
            else:
                addition_idx = np.random.choice(np.argwhere(mask == 1).reshape(-1,), 1)
            keypoint_idx = np.random.choice(a=full_keypoint_idx, size=sample_from_keypoint - 1, replace=False)
            keypoint_idx = np.concatenate((keypoint_idx, addition_idx))
        else:
            addition_idx_1 = np.random.choice(np.argwhere(mask[full_keypoint_idx] == 0).reshape(-1,), 1)
            addition_idx_2 = np.random.choice(np.argwhere(mask[full_keypoint_idx] == 1).reshape(-1,), 1)
            keypoint_idx = np.random.choice(a=full_keypoint_idx, size=sample_from_keypoint - 2, replace=False)
            keypoint_idx = np.concatenate((keypoint_idx, addition_idx_1, addition_idx_2))

    if sample_from_keypoint < int(sample_ratio * len(point_cloud)):
        addition_sample = np.random.choice(np.delete(np.arange(len(point_cloud)), full_keypoint_idx),
                                           int(sample_ratio * len(point_cloud)) - sample_from_keypoint)
        keypoint_idx = np.concatenate((keypoint_idx, addition_sample))

    return keypoint_idx, sample_from_keypoint


def draw_Point_Cloud(Points, Lables=None, axis=True, save_name = "default", **kags):

    x_axis = Points[:, 1]
    y_axis = Points[:, 2]
    z_axis = Points[:, 0]
    fig = plt.figure()
    ax = Axes3D(fig)

    ax.scatter(x_axis, y_axis, z_axis, c=Lables)
    # 设置坐标轴显示以及旋转角度
    ax.set_xlabel('x')
    ax.set_ylabel('y')
    ax.set_zlabel('z')
    ax.view_init(elev=25, azim=0)
    if not axis:
        # 关闭显示坐标轴
        plt.axis('off')

    plt.savefig( save_name + ".png")

def compute_mIoU(pred_label, gt_label):
    minl, maxl = np.min(gt_label), np.max(gt_label)
    ious = []
    for l in range(minl, maxl+1):
        I = np.sum(np.logical_and(pred_label == l, gt_label == l))
        U = np.sum(np.logical_or(pred_label == l, gt_label == l))
        if U == 0: iou = 1
        else: iou = float(I) / U
        ious.append(iou)
    return np.mean(ious)

parser = argparse.ArgumentParser()
parser.add_argument('--use_gdc', action='store_true',
                    help='Use GDC preprocessing.')
args = parser.parse_args()


save2txt = open("result_1014_key.txt", 'a+')
save2txt_log = open("logs_1014_key.txt",'a+')
ShapeNetPseudoLabeling = None
cat = 0
processed_sample = 0

import h5py

#f = h5py.File("data/shapenet_part_seg_hdf5_data/pseudo-labeling-shapnet.h5", 'w')
#f_2 = h5py.File("data/shapenet_part_seg_hdf5_data/pseudo-labeling-shapnet_cls.h5", 'w')
f = h5py.File("testing_pseudo-labeling-realword_1014.h5", 'w')
f_2 = h5py.File("testing_pseudo-labeling-realworld_cls_1014.h5", 'w')
target_names = ["bag", "bin", "box", "cabinet", "chair", "desk", "display", "door", "shelf", "table", "bed","pillow", "sink", "sofa", "toilet"]
sample_cls = []
oAcc = []
refineoAcc = []

res = open('realworldResult_1014_key.txt','a+')
prev_label = 0
count_label = 0
mious,mtrain_acc, mtest_acc = [], [], []

for seg_class in [1]: #seg_classes_keys ["Rocket"]
    catAcc = []
    refinecatAcc=[]
    
    #path = osp.join(osp.dirname(osp.realpath(__file__)), '..', 'data', 'ShapeNet')
    #pre_transform = T.NormalizeScale()
    #shapenet_dataset = ShapeNet(path, category, split='trainval', pre_transform=pre_transform)
    shapenet_dataset = RealPointCloudRaw(2048, partition='training', shuffle = False)
    print(" len " , len(shapenet_dataset))
    #shapnet_pcd = shapenet_dataset[0]
    point_features = 3
    num_classes = 2
    print("seg classes: ",num_classes)

    def train():
        model.train()
        optimizer.zero_grad()
        F.nll_loss(model()[data.train_mask], data.y[data.train_mask]).backward()
        optimizer.step()


    @torch.no_grad()
    def test():
        model.eval()
        logits, accs = model(), []
        for _, mask in data('train_mask', 'test_mask'):
            pred = logits[mask].max(1)[1]
            acc = pred.eq(data.y[mask]).sum().item() / mask.sum().item()
            accs.append(acc)
        return accs,logits.max(1)[1]

    bestIoUs = []
    for dataseti in range(len(shapenet_dataset)): #range(len(shapenet_dataset))
        pos, origilabel, seg  = shapenet_dataset[dataseti]
        if prev_label != origilabel:
            category_miou, mtrainacc, mtestacc = 0.0, 0.0, 0.0
            for i in range(len(mious)):
                category_miou+=mious[i]
                mtrainacc+=mtrain_acc[i]
                mtestacc += mtest_acc[i]
            category_miou/=len(mious)
            mtrainacc /= len(mious)
            mtestacc /= len(mious)

            res.write(target_names[prev_label] + " " + str(len(mious)) + " miou : {:.5f}  train acc : {:.5f}  test acc : {:.5f}\n".format(category_miou, mtrainacc, mtestacc))

            prev_label = origilabel
            count_label = 0
            mious = []
            mtrain_acc = []
            mtest_acc = []
        else:
            count_label += 1
        category = target_names[origilabel]
        
        seg_processed = [0 if i == -1  else 1 for i in seg]
        print(Counter(seg_processed))   # 0:background, 1:foreground
        #print(seg_processed)
        numpoints  = pos.shape[0]
        batch = torch.tensor([0] * numpoints)

        '''
        label_ratio = 0.01
        dic_label_idx = {}
        for idx, label_ in enumerate(seg_processed):
            if label_ not in dic_label_idx:
                dic_label_idx[label_] = [idx]
            else:
                dic_label_idx[label_].append(idx)

        traininglabelIndex = []
        for dic_key in dic_label_idx.keys():
            dic_key_list_len = len(dic_label_idx[dic_key])
            sample_idxs = int(dic_key_list_len * label_ratio)
            if (sample_idxs == 0):
                sample_idxs = 1

            idx2choose = np.random.choice(dic_key_list_len, sample_idxs, replace=False)

            dic_key_label = np.array(dic_label_idx[dic_key])[idx2choose]
            traininglabelIndex.extend(dic_key_label.tolist())
        '''
        '''We use keypoint detection to get points' indices to be annotated'''
        traininglabelIndex, _ = keypoint_detect(point_cloud=pos, mask=np.array(seg_processed), sample_ratio=0.01)

        #print(traininglabelIndex)
        #print("trainging len ", len(traininglabelIndex))
        #print(len(idx2choose))
        train_mask, test_mask ,val_mask = [],[],[]
        for i in range(numpoints):
            if i in traininglabelIndex:
                train_mask.append(True)
                test_mask.append(False)
            else:
                train_mask.append(False)
                test_mask.append(True)
            val_mask.append(False)

        print("train mask count: " , Counter(train_mask)[True])
        print("test_mask mask count: " , Counter(test_mask)[True])

    ##################build  data###################


        mydata = Data(x = torch.tensor(pos))
        mydata.train_mask = torch.tensor(train_mask).bool()
        #print(Counter(mydata.train_mask.numpy().tolist()))


        mydata.test_mask = torch.tensor(test_mask).bool()

        #print(Counter(mydata.test_mask.numpy().tolist()))
        mydata.val_mask = torch.tensor(val_mask).bool()
        mydata.y = torch.tensor(seg_processed).long()
        data = mydata
        # device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        device = 'cuda'
        data = data.to(device)
        batch = batch.to(device)
        best_val_acc = test_acc = 0
        epoch = 64
        best_miou = 0.0
        k = 20
        # rgcn
        model = RelationGraphConvNet(numpoints,num_bases = 30).to(device)
        # gcn
        #model = GCN().to(device)
        optimizer = torch.optim.Adam(model.parameters(), lr=0.0135)
       #optimizer = torch.optim.Adam([dict(params=model.conv1.parameters(), weight_decay=5e-4), dict(params=model.conv2.parameters(), weight_decay=5e-4)], lr=0.0135)  # Only perform weight-decay on first convolution.
        #[dict(params=model.reg_params, weight_decay=5e-4),
        #dict(params=model.non_reg_params, weight_decay=5e-4)]

        #optimizer = torch.optim.SGD(model.parameters(), lr=0.01, momentum=0.9, weight_decay=1e-4)
        scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(optimizer, T_max = epoch, eta_min=1e-4)
        #scheduler = torch.optim.lr_scheduler.StepLR(optimizer, step_size=100, gamma=0.5)  # 每十次迭代，学习率减半

        # #########################only knn k = 20 #########################
        #

        #edge_index = knn_graph(data.x, k, batch, loop=False).to(device)
        ### near and far as two kind of type
        #edge_weight = torch.tensor(([0]*k) * numpoints).long().to(device)
        # #########################only knn k = 40 #########################
        #
        #k = 20
        #edge_index = knn_graph(data.x, k, batch, loop=False).to(device)
        ### near and far as two kind of type
        #edge_weight = torch.tensor(([0]*int(k / 2) + [1]*int(k / 2)) * numpoints).long().to(device)


        #################
        x_, idx_eu, idx_ei = eigen_Graph(data.x.unsqueeze(0), k, device=device)
        edge_index = torch.cat([idx_ei, idx_eu], dim=1)
        edge_weight_eu = [0] * len(idx_eu[0])
        edge_weight_ei = [1] * len(idx_ei[0])
        edge_weight = torch.tensor(edge_weight_eu + (edge_weight_ei) ).long().to(device)

        # #####radius + ei graph
        # x_, idx_eu, idx_ei = eigen_Graph(data.x.unsqueeze(0), k, device=device)
        # idx_radius = radius_graph(data.x, r = 0.1, batch = batch)
        # edge_index = torch.cat([idx_ei, idx_radius], dim=1)
        # edge_weight_ei = [0] * len(idx_ei[0])
        # edge_weight_er = [1] * len(idx_radius[0])
        # edge_weight = torch.tensor(edge_weight_ei + edge_weight_er).long().to(device)
        #k = 40
        #edge_index = knn_graph(data.x, k, batch, loop=False).to(device)
        
        #####################################
   
        
        # ############ radius + knn + far nn
        # k = 20
        # edge_index_knn = knn_graph(data.x, k*2, batch, loop=False).to(device)
        # x_, idx_eu, idx_ei = eigen_Graph(data.x.unsqueeze(0), k, device=device)
        # select_mask =  torch.tensor([[False]*k + [True]*k]*numpoints).bool().to(device).reshape(-1)
        # edge_index1 = torch.masked_select(edge_index_knn[0], select_mask)
        # edge_index2 = torch.masked_select(edge_index_knn[1], select_mask)
        # edge_index_fnn = torch.cat([edge_index1.reshape(1,-1),edge_index2.reshape(1,-1)],dim=0)

        # edge_index1 = torch.masked_select(edge_index_knn[0], ~select_mask)
        # edge_index2 = torch.masked_select(edge_index_knn[1], ~select_mask)
        # edge_index_knn = torch.cat([edge_index1.reshape(1, -1), edge_index2.reshape(1, -1)], dim=0)

        # idx_radius = radius_graph(data.x, r = 0.05, batch = batch)
        # edge_index = torch.cat([edge_index_knn, edge_index_fnn, idx_radius], dim=1)
        # edge_weight_knn = [0] * len(edge_index_knn[0])
        # edge_weight_farnn = [1] * len(edge_index_fnn[0])
        # edge_weight_er = [2] * len(idx_radius[0])
 
        # edge_weight = torch.tensor(edge_weight_knn + edge_weight_farnn +  edge_weight_er).long().to(device)

        #data.x = x_.squeeze(0)[:,:15]. contiguous()
        #print(data.x.shape)÷
        #print(x_.shape)
        #datax = torch.cat([data.x, x_.squeeze(0)], dim = 1)
        print(data.x.shape)
        #input()
        trainAccs = []
        testAccs = []
        debug = False
        perfect_train_acc = 0
        perfect_train_accs = []
        for epoch in tqdm(range(1, epoch+1),ascii = True):

            train()
            scheduler.step()  # 学习率迭代次数+1
            accs, pred = test()
            train_acc, tmp_test_acc = accs[0], accs[1]
            if(train_acc >= 0.99):
                perfect_train_acc+=1
                perfect_train_accs.append(perfect_train_acc)
            if(perfect_train_acc == 3):
               break
            if debug :
                accs, pred = test()
                train_acc, tmp_test_acc = accs[0], accs[1]
                trainAccs.append(train_acc)
                testAccs.append(tmp_test_acc)
        #best_test_acc_index = np.argmax(testAccs)
        #print("best at " + str(best_test_acc_index) + " perfect times: " + str(perfect_train_accs[best_test_acc_index]) + "best acc " + str(testAccs[best_test_acc_index]))
        if debug :
            import matplotlib as mpl
            import matplotlib.pyplot as plt
            x_axix = list(range(1,epoch+1))
            plt.title('Result Analysis')
            plt.plot(x_axix, trainAccs, color='green', label='training accuracy')
            plt.plot(x_axix, testAccs, color='red', label='testing accuracy')
            plt.legend() # 显示图例
            plt.xlabel('iteration times')
            plt.ylabel('rate')
            plt.show()
            x=input()
        accs, pred = test()
        train_acc, tmp_test_acc = accs[0], accs[1]
        acc_log = (" {} train acc: {:.5f} test acc: {:.5f}").format(dataseti, train_acc,tmp_test_acc)
        print(acc_log)
        curr_miou = compute_mIoU(pred.cpu().numpy(), data.y.cpu().numpy())
        print("mIoU : ", curr_miou)
        save2txt_log.write(acc_log)
        save2txt_log.write(" mIoU : {:.5f} \n".format(curr_miou))
        bestIoUs.append(curr_miou)
        catAcc.append(tmp_test_acc)
        for idx in traininglabelIndex:
            pred[idx] = data.y[idx]
        curr_miou = compute_mIoU(pred.cpu().numpy(), data.y.cpu().numpy())
        print("new miou:", curr_miou)
        mious.append(curr_miou)
        mtrain_acc.append(train_acc)
        mtest_acc.append(tmp_test_acc)
        # # resample
        # numpoints_to_resample = 2048
        # choice = np.random.choice(len(pred), numpoints_to_resample, replace=True)
        # points = data.x.cpu().numpy()[choice, : ]
        # seg = pred.cpu().numpy()[choice].reshape(-1,1)
        # data2save = np.concatenate([points, seg], axis = 1).reshape(1,numpoints_to_resample,-1)

        #without resample
        points = data.x.cpu().numpy()
        predseg = pred.cpu().numpy().reshape(-1,1) #+ seg_classes[category][0]
        #print(seg)
        mask2save = mydata.train_mask.cpu().numpy().reshape(-1,1)
        #print(mask2save[:,0])
        #print(Counter(mask2save[:,0].tolist()))
       

        #np.savetxt("wrong_labeled_data.txt",np.concatenate([points,np.abs(predseg-seg.reshape(-1,1))],axis = 1))

        #np.savetxt("all.txt",data2save)
        
        # voting module
        idx_eu = knn_graph(data.x, 50, batch, loop=True).to(device)
        #[N*k]
        neighborPred1 = predseg[idx_eu[0].cpu()].reshape(numpoints,-1)
        #neighborPred2 = predseg[idx_ei[0].cpu()].reshape(numpoints,-1)
        #print(neighborPred1.shape)

        #neighborPred = np.concatenate([neighborPred1,neighborPred2], axis = 1)
        #print(neighborPred1.shape)
        #[N*k] -> [N,K] ->[N]
        marjortiyPred1 = mode(neighborPred1, axis = 1)[0].ravel()
        #marjortiyPred2 = mode(neighborPred2, axis = 1)[0].ravel()
        #print(marjortiyPred)
        #print(marjortiyPred.shape)
        # if marjorityPre != predseg, down sample!
        #print(predseg.shape)
        #print(predseg.reshape(-1)[:100])
        #print(marjortiyPred1[:100])
        downSampleMask = (predseg.reshape(-1) == ( marjortiyPred1))
        #print(downSampleMask[:100])
        #np.savetxt("marjortiyPredSample.txt",np.concatenate([points,downSampleMask.reshape(-1,1)],axis = 1))
        #computing new acc
        #train mask acc = 1
        #predseg = predseg.reshape(-1)
        #seg = seg.reshape(-1)
        #print(predseg == seg)
        #print((predseg == seg).sum())
        #mask2save = mydata.train_mask.cpu().numpy().reshape(-1)
        #print(mask2save.sum())
        #print(mask2save)
        #print(predseg[mask2save])
        #print(predseg[mask2save] == seg[mask2save])
        #print((predseg[mask2save] == seg[mask2save]).sum())
        
       
        #print(predseg[(~mask2save) & downSampleMask] == seg[(~mask2save) & downSampleMask])
        #print((predseg[loss_on_noisy_labels_mask] == seg[loss_on_noisy_labels_mask]).sum())
        
        #print(((predseg[mask2save] == seg[mask2save]).sum() + (predseg[loss_on_noisy_labels_mask] == seg[loss_on_noisy_labels_mask]).sum() ) / (mask2save.sum() + (loss_on_noisy_labels_mask).sum())
        #x = input()
        #continue;
        seg_label = np.max(seg)
        predseg = np.array([-1 if i == 0  else seg_label for i in predseg])
        #seg = np.array([0 if i == -1  else seg_label for i in seg])
        data2save = np.concatenate([points, predseg.reshape(-1,1),seg.reshape(-1,1), mask2save.reshape(-1,1),downSampleMask.reshape(-1,1)], axis=1)
        """
        print(data2save[:,3])
        print(data2save[:,4])
        print(Counter(predseg))
        print(Counter(seg))
        print(Counter(mask2save.reshape(-1).tolist()))
        print(Counter(downSampleMask.reshape(-1).tolist()))
        """
        f.create_dataset(str(processed_sample).zfill(5), data = data2save)

        sample_cls.append(origilabel)
        processed_sample += 1
        seg = np.array(seg)
        # if(ShapeNetPseudoLabeling == None):
        #     ShapeNetPseudoLabeling = data2save.copy()
        # else:
        #     ShapeNetPseudoLabeling = np.concatenate(ShapeNetPseudoLabeling, data2save)
        #### save data: pos, normal, seg, cls
        
        
        predseg = predseg.reshape(-1)
        seg = seg.reshape(-1)
        loss_on_noisy_labels_mask = (~mask2save.reshape(-1)) & (downSampleMask)
        mask2save = mask2save.reshape(-1)
        refineAcc = ((predseg[mask2save] == seg[mask2save]).sum() + (predseg[loss_on_noisy_labels_mask] == seg[loss_on_noisy_labels_mask]).sum() ) / (mask2save.sum() + (loss_on_noisy_labels_mask).sum())
       
        refinecatAcc.append(refineAcc)
        print("refine new acc : ", refineAcc)

    cat += 1
    print(bestIoUs)
    sum = 0.0
    i = 0
    for iou in bestIoUs:
        sum += iou
        i+=1
    print(" mean IoU : " , sum/i)
    save2txt.write(category + " : {:.5f} \n".format(sum/i))
    #save2txt.write(str(i) + "\n")
    sum = 0.0
    for acc in catAcc:
        sum+=acc
    sum/=len(catAcc)
    oAcc.append(sum)
    
    sum = 0.0
    for acc in refinecatAcc:
        sum += acc
    sum /= len(refinecatAcc)
    refineoAcc.append(sum)
    
    
sum1 = 0.0
for acc in oAcc:
    sum1+=acc
sum1/=len(oAcc)

sum2 = 0.0
for acc in refineoAcc:
    sum2+=acc
sum2/=len(refineoAcc)
s = "cat acc : " + str(oAcc) + " oAcc: " + str(sum1) + "\n"
s += "refine cat acc : " + str(refineoAcc) + "refine oAcc: " + str(sum2) + "\n" 

save2txt.write(s)

category_miou, mtrainacc, mtestacc = 0.0, 0.0, 0.0
for i in range(len(mious)):
    category_miou += mious[i]
    mtrainacc += mtrain_acc[i]
    mtestacc += mtest_acc[i]
category_miou /= len(mious)
mtrainacc /= len(mious)
mtestacc /= len(mious)

res.write(
    target_names[prev_label] + " " + str(len(mious)) + " miou : {:.5f}  train acc : {:.5f}  test acc : {:.5f}\n".format(category_miou, mtrainacc,
                                                                                               mtestacc))
res.close()
f_2.create_dataset("cls", data = np.array(sample_cls))
#
# ff = h5py.File("pseudo-labeling-shapnet.h5", 'r')
# for keys in ff.keys():
#     print(ff[keys].shape)
# ff_2 = h5py.File("pseudo-labeling-shapnet_cls.h5",'r')
# print(ff_2['cls'])

f.close()
f_2.close()

# ff.close()
# ff_2.close()
# print(ShapeNetPseudoLabeling)
# label2sava = torch.from_numpy(np.array(ShapeNetPseudoLabeling))
# print(label2sava)
save2txt.close()
