import os.path as osp
import argparse

import torch
import torch.nn.functional as F
from torch_geometric.datasets import ShapeNet
from torch_geometric.data import Data
import torch_geometric.transforms as T
#from rgcn_conv import RGCNConv
from torch_geometric.nn import GCNConv, DynamicEdgeConv,knn_graph, EdgeConv, radius_graph, RGCNConv# noqa
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import random
#import h5py
import numpy as np
from collections import Counter
from tqdm import tqdm
from EigenGraph import eigen_Graph
from data import ShapeNetPart

random.seed(100)
torch.manual_seed(7)
np.random.seed(100)
from torch.nn import Sequential as Seq, Dropout, Linear as Lin,BatchNorm1d as BN
def MLP(channels, bias=True):
    return Seq(*[
        Seq(Lin(channels[i - 1], channels[i], bias=bias),BN(channels[i]),torch.nn.LeakyReLU(negative_slope=0.2),Lin(channels[i], channels[i], bias=bias))
        for i in range(1, len(channels))
    ])
class GCN(torch.nn.Module):
    def __init__(self):
        super(GCN, self).__init__()
        self.conv1 = GCNConv(point_features, 16, cached=True)
        self.conv2 = GCNConv(16, num_classes, cached=True)
        # self.conv1 = ChebConv(data.num_features, 16, K=2)
        # self.conv2 = ChebConv(16, data.num_features, K=2)

    def forward(self):
        x= data.x
        x = F.relu(self.conv1(x, edge_index))
        x = F.dropout(x, training=self.training)
        x = self.conv2(x, edge_index)
        return F.log_softmax(x, dim=1)


class RelationGraphConvNet(torch.nn.Module):
    def __init__(self,num_points, k = 20,num_bases = 5):
        super(RelationGraphConvNet, self).__init__()
        self.conv1 = RGCNConv(point_features, 16, num_relations = 4, num_bases = num_bases)
        self.conv2 = RGCNConv(16, num_classes, num_relations = 4, num_bases = num_bases)
        self.reg_params = self.conv1.parameters()
        self.non_reg_params = self.conv2.parameters()
        self.k = k
        self.num_points = num_points;

    def forward(self):
        x = data.x
        ##############r-gcn
        x = F.relu(self.conv1(x, edge_index, edge_weight))
        x = F.dropout(x, training=self.training)
        x = self.conv2(x, edge_index, edge_weight)

        return F.log_softmax(x, dim=1)




def draw_Point_Cloud(Points, Lables=None, axis=True, save_name = "default", **kags):

    x_axis = Points[:, 1]
    y_axis = Points[:, 2]
    z_axis = Points[:, 0]
    fig = plt.figure()
    ax = Axes3D(fig)

    ax.scatter(x_axis, y_axis, z_axis, c=Lables)
    # 设置坐标轴显示以及旋转角度
    ax.set_xlabel('x')
    ax.set_ylabel('y')
    ax.set_zlabel('z')
    ax.view_init(elev=25, azim=0)
    if not axis:
        # 关闭显示坐标轴
        plt.axis('off')

    plt.savefig( save_name + ".png")

def compute_mIoU(pred_label, gt_label):
    minl, maxl = np.min(gt_label), np.max(gt_label)
    ious = []
    for l in range(minl, maxl+1):
        I = np.sum(np.logical_and(pred_label == l, gt_label == l))
        U = np.sum(np.logical_or(pred_label == l, gt_label == l))
        if U == 0: iou = 1
        else: iou = float(I) / U
        ious.append(iou)
    return np.mean(ious)
seg_classes = {
        'airplane': [0, 1, 2, 3],
        'bag': [4, 5],
        'cap': [6, 7],
        'car': [8, 9, 10, 11],
        'chair': [12, 13, 14, 15],
        'earphone': [16, 17, 18],
        'guitar': [19, 20, 21],
        'knife': [22, 23],
        'lamp': [24, 25, 26, 27],
        'laptop': [28, 29],
        'motorbike': [30, 31, 32, 33, 34, 35],
        'mug': [36, 37],
        'pistol': [38, 39, 40],
        'rocket': [41, 42, 43],
        'skateboard': [44, 45, 46],
        'table': [47, 48, 49],
    }
parser = argparse.ArgumentParser()
parser.add_argument('--use_gdc', action='store_true',
                    help='Use GDC preprocessing.')
parser.add_argument('--exp',help='expname.')
args = parser.parse_args()

seg_classes_keys = seg_classes.keys()
save2txt = open("result.txt", 'a+')
save2txt_log = open("logs.txt",'a+')
ShapeNetPseudoLabeling = None
cat = 0
processed_sample = 0
import h5py
f = h5py.File("pseudo-labeling-shapnet.h5", 'w')
f_2 = h5py.File("pseudo-labeling-shapnet_cls.h5", 'w')
sample_cls = []
oAcc = []
print(args.exp)
for seg_class in ["airplane"]: #seg_classes_keys ["Rocket"]
    catAcc = []
    category = seg_class.lower() #'Chair'
    #path = osp.join(osp.dirname(osp.realpath(__file__)), '..', 'data', 'ShapeNet')
    #pre_transform = T.NormalizeScale()
    #shapenet_dataset = ShapeNet(path, category, split='trainval', pre_transform=pre_transform)
    shapenet_dataset = ShapeNetPart(partition='test', num_points=2048, class_choice=category)
    print(category, " len " , len(shapenet_dataset))
    #shapnet_pcd = shapenet_dataset[0]
    point_features = 3
    num_classes = seg_classes[category][-1] - seg_classes[category][0] + 1
    print("seg classes: ",num_classes)

    def train():
        model.train()
        optimizer.zero_grad()
        F.nll_loss(model()[data.train_mask], data.y[data.train_mask]).backward()
        optimizer.step()


    @torch.no_grad()
    def test():
        model.eval()
        logits, accs = model(), []
        for _, mask in data('train_mask', 'test_mask'):
            pred = logits[mask].max(1)[1]
            acc = pred.eq(data.y[mask]).sum().item() / mask.sum().item()
            accs.append(acc)
        return accs,logits.max(1)[1]

    bestIoUs = []
    for dataseti in [0,1]:#range(len(shapenet_dataset)): #
        pos, label, seg  = shapenet_dataset[dataseti]
        seg_processed = seg - seg_classes[category][0]
        print(seg_processed)
        numpoints  = pos.shape[0]

        k = 20
        label_ratio = 0.1
        batch = torch.tensor([0]*numpoints)

        dic_label_idx = {}
        for idx, label in enumerate(seg_processed):
            if label not in dic_label_idx:
                dic_label_idx[label] = [idx]
            else:
                dic_label_idx[label].append(idx)

        traininglabelIndex = []
        for dic_key in dic_label_idx.keys():
            dic_key_list_len = len(dic_label_idx[dic_key])
            sample_idxs = int(dic_key_list_len * label_ratio)
            if (sample_idxs == 0):
                sample_idxs = 1
            idx2choose = np.random.choice(dic_key_list_len, sample_idxs, replace=False)
            dic_key_label = np.array(dic_label_idx[dic_key])[idx2choose]
            traininglabelIndex.extend(dic_key_label.tolist())

        train_mask, test_mask ,val_mask = [],[],[]
        for i in range(numpoints):
            if i in traininglabelIndex:
                train_mask.append(True)
                test_mask.append(False)
            else:
                train_mask.append(False)
                test_mask.append(True)
            val_mask.append(False)

        print("train mask count: " , Counter(train_mask)[True])
        print("test_mask mask count: " , Counter(test_mask)[True])

    ##################build  data###################


        mydata = Data(x = torch.tensor(pos))
        mydata.train_mask = torch.tensor(train_mask).bool();
        mydata.test_mask = torch.tensor(test_mask).bool();
        mydata.val_mask = torch.tensor(val_mask).bool();
        mydata.y = torch.tensor(seg_processed).long();
        data = mydata
        max_labels = np.max(seg_processed)+1
        weaklabels = [label if mask else max_labels for label, mask in zip(seg_processed, train_mask)]
        wrrite_labels = np.concatenate([pos,seg_processed.reshape(-1,1),np.array(weaklabels).reshape(-1,1)], axis=1)
        np.savetxt("weaklylabels.txt",wrrite_labels)
        x = input()
        trainlabel = mydata.y.cpu().numpy().copy()
        for i in range(len(trainlabel)):
            if(train_mask[i] == False):
                trainlabel[i] = 6;
        pos = mydata.x.cpu().numpy()
        gtlabel = mydata.y.cpu().numpy()
        outdata = np.concatenate([pos.reshape(-1,3), trainlabel.reshape(-1,1), gtlabel.reshape(-1,1)], axis=1)
        print(outdata.shape)
        np.savetxt("weaklydata.txt", outdata)
        x = input()

        # device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        device = 'cuda'
        data = data.to(device)
        batch = batch.to(device)
        best_val_acc = test_acc = 0
        epoch = 250
        best_miou = 0.0
        # gcn
        # model = GCN().to(device)
        # rgcn
        model = RelationGraphConvNet(numpoints,num_bases = 10).to(device)
        optimizer = torch.optim.Adam(model.parameters(), lr=0.0135)
        # [dict(params=model.reg_params, weight_decay=5e-4),
        # dict(params=model.non_reg_params, weight_decay=5e-4)]

        # optimizer = torch.optim.SGD([ dict(params=model.reg_params, weight_decay=5e-4),
        #                               dict(params=model.non_reg_params, weight_decay=5e-4)],
        #                             lr=0.02)
        #scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(optimizer, T_max=epoch, eta_min=1e-4)
        #scheduler = torch.optim.lr_scheduler.StepLR(optimizer, step_size=100, gamma=0.5)  # 每十次迭代，学习率减半
        # #########################set = 3 knn = 40， 20-> 0 20 ->1 #########################
        #
        # k = 40 near+far knn
        # k = 40
        # edge_index = knn_graph(data.x, k, batch, loop=False).to(device)
        # edge_weight = torch.tensor(([0]*int(k / 2) + [1]*int(k / 2)) * numpoints).long().to(device)


        ################# radius + eu + ei graph
        # x_, idx_eu, idx_ei = eigen_Graph(data.x.unsqueeze(0), k, device=device)
        # idx_radius = radius_graph(data.x, r = 0.1, batch = batch)
        # edge_index = torch.cat([idx_ei, idx_eu, idx_radius], dim=1)
        # edge_weight_ei = [0] * len(idx_ei[0])
        # edge_weight_eu = [1] * len(idx_eu[0])
        # edge_weight_er = [2] * len(idx_radius[0])
        # edge_weight = torch.tensor(edge_weight_eu + (edge_weight_ei) + edge_weight_er).long().to(device)

        # #####radius + ei graph
        # x_, idx_eu, idx_ei = eigen_Graph(data.x.unsqueeze(0), k, device=device)
        # idx_radius = radius_graph(data.x, r = 0.1, batch = batch)
        # edge_index = torch.cat([idx_ei, idx_radius], dim=1)
        # edge_weight_ei = [0] * len(idx_ei[0])
        # edge_weight_er = [1] * len(idx_radius[0])
        # edge_weight = torch.tensor(edge_weight_ei + edge_weight_er).long().to(device)
        # ############ radius + far nn

        # edge_index_knn = knn_graph(data.x, k*2, batch, loop=False).to(device)
        # select_mask =  torch.tensor([[False]*k + [True]*k]*numpoints).bool().to(device).reshape(-1)
        # edge_index1 = torch.masked_select(edge_index_knn[0], select_mask)
        # edge_index2 = torch.masked_select(edge_index_knn[1], select_mask)
        # edge_index_fnn = torch.cat([edge_index1.reshape(1,-1),edge_index2.reshape(1,-1)],dim=0)
        #
        # idx_radius = radius_graph(data.x, r = 0.1, batch = batch)
        # edge_index = torch.cat([edge_index_fnn, idx_radius], dim=1)
        # edge_weight_ei = [0] * len(edge_index_fnn[0])
        # edge_weight_er = [1] * len(idx_radius[0])
        # edge_weight = torch.tensor(edge_weight_ei + edge_weight_er).long().to(device)
        ############ours
        # x_, idx_eu, idx_ei = eigen_Graph(data.x.unsqueeze(0), k, device=device)
        #
        # edge_index = torch.cat([idx_ei, idx_eu], dim=1)
        # edge_weight_eu = [0] * len(idx_eu[0])
        # edge_weight_ei = [1] * len(idx_ei[0])
        # edge_weight = torch.tensor(edge_weight_eu + (edge_weight_ei) ).long().to(device)
        # ############ radius + knn + far nn

        edge_index_knn = knn_graph(data.x, k*2, batch, loop=False).to(device)
        x_, idx_eu, idx_ei = eigen_Graph(data.x.unsqueeze(0), k, device=device)
        select_mask =  torch.tensor([[False]*k + [True]*k]*numpoints).bool().to(device).reshape(-1)
        edge_index1 = torch.masked_select(edge_index_knn[0], select_mask)
        edge_index2 = torch.masked_select(edge_index_knn[1], select_mask)
        edge_index_fnn = torch.cat([edge_index1.reshape(1,-1),edge_index2.reshape(1,-1)],dim=0)

        edge_index1 = torch.masked_select(edge_index_knn[0], ~select_mask)
        edge_index2 = torch.masked_select(edge_index_knn[1], ~select_mask)
        edge_index_knn = torch.cat([edge_index1.reshape(1, -1), edge_index2.reshape(1, -1)], dim=0)

        idx_radius = radius_graph(data.x, r = 0.1, batch = batch)
        edge_index = torch.cat([edge_index_knn, edge_index_fnn, idx_radius, idx_ei], dim=1)
        edge_weight_knn = [0] * len(edge_index_knn[0])
        edge_weight_farnn = [1] * len(edge_index_fnn[0])
        edge_weight_er = [2] * len(idx_radius[0])
        edge_weight_ei = [3] * len(idx_ei[0])
        edge_weight = torch.tensor(edge_weight_knn +edge_weight_farnn +  edge_weight_er+edge_weight_ei).long().to(device)

        for epoch in tqdm(range(1, epoch+1),ascii = True):

            train()
            #scheduler.step()  # 学习率迭代次数+1

        accs, pred = test()
        train_acc, tmp_test_acc = accs[0], accs[1]
        acc_log = category + (" {} train acc: {:.5f} test acc: {:.5f}").format(dataseti, train_acc,tmp_test_acc)
        print(acc_log)
        curr_miou = compute_mIoU(pred.cpu().numpy(), data.y.cpu().numpy())
        print("mIoU : ", curr_miou)
        save2txt_log.write(acc_log)
        save2txt_log.write(" mIoU : {:.5f} \n".format(curr_miou))
        bestIoUs.append(curr_miou)
        catAcc.append(tmp_test_acc)
        for idx in idx2choose:
            pred[idx] = data.y[idx];
        curr_miou = compute_mIoU(pred.cpu().numpy(), data.y.cpu().numpy())
        print("new miou:", curr_miou)
        # # resample
        # numpoints_to_resample = 2048
        # choice = np.random.choice(len(pred), numpoints_to_resample, replace=True)
        # points = data.x.cpu().numpy()[choice, : ]
        # seg = pred.cpu().numpy()[choice].reshape(-1,1)
        # data2save = np.concatenate([points, seg], axis = 1).reshape(1,numpoints_to_resample,-1)

        #without resample
        points = data.x.cpu().numpy()
        seg = pred.cpu().numpy().reshape(-1,1)+ seg_classes[category][0]
        print(seg)
        data2save = np.concatenate([points, seg], axis=1)
        f.create_dataset(str(processed_sample).zfill(5), data = data2save)
        sample_cls.append(cat)
        # if(ShapeNetPseudoLabeling == None):
        #     ShapeNetPseudoLabeling = data2save.copy()
        # else:
        #     ShapeNetPseudoLabeling = np.concatenate(ShapeNetPseudoLabeling, data2save)
        #### save data: pos, normal, seg, cls
        processed_sample += 1

    cat += 1
    print(bestIoUs)
    sum = 0.0
    i = 0
    for iou in bestIoUs:
        sum += iou
        i+=1
    print(category, " mean IoU : " , sum/i)
    save2txt.write(category + " : {:.5f} \n".format(sum/i))
    #save2txt.write(str(i) + "\n")
    sum = 0.0
    for acc in catAcc:
        sum+=acc
    sum/=len(catAcc)
    oAcc.append(sum)
sum = 0.0
for acc in oAcc:
    sum+=acc
sum/=len(oAcc)
s = "cat acc : " + str(oAcc) + " oAcc: " + str(sum)+"\n"
save2txt.write(s)
save2txt.write(str(args.exp)+"\n")
f_2.create_dataset("cls", data = np.array(sample_cls))
#
# ff = h5py.File("pseudo-labeling-shapnet.h5", 'r')
# for keys in ff.keys():
#     print(ff[keys].shape)
# ff_2 = h5py.File("pseudo-labeling-shapnet_cls.h5",'r')
# print(ff_2['cls'])

f.close()
f_2.close()

# ff.close()
# ff_2.close()
# print(ShapeNetPseudoLabeling)
# label2sava = torch.from_numpy(np.array(ShapeNetPseudoLabeling))
# print(label2sava)
save2txt.close()
