import argparse
import math
import h5py
import numpy as np
import tensorflow as tf
#import tensorflow.compat.v1 as tf
#tf.disable_v2_behavior()
import socket
import importlib
import os
import sys
BASE_DIR = os.path.dirname(os.path.abspath(__file__))
ROOT_DIR = BASE_DIR
sys.path.append(BASE_DIR)
sys.path.append(os.path.join(BASE_DIR, 'models'))
sys.path.append(os.path.join(BASE_DIR, 'utils'))
import provider
from utils import tf_util
sys.path.append(os.path.join(BASE_DIR, '..'))
import data_utils
from tqdm import tqdm
parser = argparse.ArgumentParser()
parser.add_argument('--gpu', type=int, default=0, help='GPU to use [default: GPU 0]')
parser.add_argument('--model', default='dgcnn_bga', help='Model name: dgcnn')

parser.add_argument('--log_dir', default='log/', help='Log dir [default: log]')
parser.add_argument('--with_bg', default = True, help='Whether to have background or not [default: True]')
parser.add_argument('--norm', default = True, help='Whether to normalize data or not [default: False]')
parser.add_argument('--center_data', default = True, help='Whether to explicitly center the data [default: False]')
parser.add_argument('--seg_weight', type=int, default=0.5, help='Segmentation weight in loss')
parser.add_argument('--num_class', type=int, default = 15, help='Number of classes to classify.')
#training_objectdataset_augmentedrot_scale75
parser.add_argument('--exp_type', default = '1', help='10% or 1%')
parser.add_argument('--train_file', default = 'model_data/1%_train_processed.h5', help='Location of test file')
parser.add_argument('--test_file', default = 'model_data/test_objectdataset_augmentedrot_scale75.h5', help='Location of test file')

parser.add_argument('--num_point', type=int, default=1024, help='Point Number [256/512/1024/2048] [default: 1024]')
parser.add_argument('--max_epoch', type=int, default=250, help='Epoch to run [default: 250]')
parser.add_argument('--batch_size', type=int, default=32, help='Batch Size during training [default: 32]')
parser.add_argument('--learning_rate', type=float, default=0.001, help='Initial learning rate [default: 0.001]')
parser.add_argument('--momentum', type=float, default=0.9, help='Initial learning rate [default: 0.9]')
parser.add_argument('--optimizer', default='adam', help='adam or momentum [default: adam]')
parser.add_argument('--decay_step', type=int, default=200000, help='Decay step for lr decay [default: 200000]')
parser.add_argument('--decay_rate', type=float, default=0.7, help='Decay rate for lr decay [default: 0.8]')
FLAGS = parser.parse_args()

DATA_DIR = os.path.join(ROOT_DIR, '../../../../')
BATCH_SIZE = FLAGS.batch_size
NUM_POINT = FLAGS.num_point
MAX_EPOCH = FLAGS.max_epoch
BASE_LEARNING_RATE = FLAGS.learning_rate
GPU_INDEX = FLAGS.gpu
MOMENTUM = FLAGS.momentum
OPTIMIZER = FLAGS.optimizer
DECAY_STEP = FLAGS.decay_step
DECAY_RATE = FLAGS.decay_rate

WITH_BG = FLAGS.with_bg
NORMALIZED = FLAGS.norm
# if FLAGS.exp_type == '1':
#     TRAIN_FILE = '../../Realworld/1%_train_processed.h5'
# elif FLAGS.exp_type == '10':
#     TRAIN_FILE = '../../Realworld/train_processed.h5'
TRAIN_FILE = FLAGS.train_file
print("use ", TRAIN_FILE)
TEST_FILE = FLAGS.test_file
CENTER_DATA = FLAGS.center_data
SEG_WEIGHT = FLAGS.seg_weight
# global LOSS_CLS_1#tf.Variable(1.0)
# global LOSS_CLS_2  #tf.Variable(1.0)
# global LOSS_SEG_1#tf.Variable(1.0)
# global LOSS_SEG_2 #tf.Variable(1.0)
# LOSS_CLS_1 = 1.0
# LOSS_SEG_1 = 1.0
# LOSS_CLS_2 = 1.0
# LOSS_SEG_2 = 1.0
MODEL = importlib.import_module(FLAGS.model) # import network module
MODEL_FILE = os.path.join(BASE_DIR, 'models', FLAGS.model+'.py')
LOG_DIR = os.path.join("checkpoints", FLAGS.log_dir)
if not os.path.exists(LOG_DIR): os.mkdir(LOG_DIR)
os.system('cp models/dgcnn_bga.py %s' % (LOG_DIR)) # bkp of model def
os.system('cp train_seg.py %s' % (LOG_DIR)) # bkp of train procedure
os.system('cp ../data_utils.py %s' % (LOG_DIR)) # bkp of data utils
LOG_FOUT = open(os.path.join(LOG_DIR, 'log_train.txt'), 'w')
LOG_FOUT.write(str(FLAGS)+'\n')
print("USE GOOGLE DRIVE VERSION")
BN_INIT_DECAY = 0.5
BN_DECAY_DECAY_RATE = 0.5
BN_DECAY_DECAY_STEP = float(DECAY_STEP)
BN_DECAY_CLIP = 0.99



HOSTNAME = socket.gethostname()

NUM_CLASSES = FLAGS.num_class

print("Number of Classes: "+str(NUM_CLASSES))
print("Normalized: "+str(NORMALIZED))
print("Center Data: "+str(CENTER_DATA))

TRAIN_DATA, TRAIN_LABELS, TRAIN_MASKS, _, Seg_train_mask, Seg_select_mask = data_utils.myload_withmask_h5(TRAIN_FILE)

TEST_DATA, TEST_LABELS, TEST_MASKS = data_utils.load_withmask_h5(TEST_FILE)      
TRAIN_MASKS = data_utils.convert_to_binary_mask(TRAIN_MASKS)
TEST_MASKS = data_utils.convert_to_binary_mask(TEST_MASKS)

if (CENTER_DATA):
    TRAIN_DATA = data_utils.center_data(TRAIN_DATA)
    TEST_DATA = data_utils.center_data(TEST_DATA)

if (NORMALIZED):
    TRAIN_DATA = data_utils.normalize_data(TRAIN_DATA)
    TEST_DATA = data_utils.normalize_data(TEST_DATA)

print(len(TRAIN_DATA))
print(len(TEST_DATA))

def log_string(out_str):
    LOG_FOUT.write(out_str+'\n')
    LOG_FOUT.flush()
    print(out_str)


def get_learning_rate(batch):
    learning_rate = tf.train.exponential_decay(
                        BASE_LEARNING_RATE,  # Base learning rate.
                        batch * BATCH_SIZE,  # Current index into the dataset.
                        DECAY_STEP,          # Decay step.
                        DECAY_RATE,          # Decay rate.
                        staircase=True)
    learning_rate = tf.maximum(learning_rate, 0.00001) # CLIP THE LEARNING RATE!
    return learning_rate        

def get_bn_decay(batch):
    bn_momentum = tf.train.exponential_decay(
                      BN_INIT_DECAY,
                      batch*BATCH_SIZE,
                      BN_DECAY_DECAY_STEP,
                      BN_DECAY_DECAY_RATE,
                      staircase=True)
    bn_decay = tf.minimum(BN_DECAY_CLIP, 1 - bn_momentum)
    return bn_decay

def get_pseudo_aplha(batch):
    bn_momentum = tf.train.exponential_decay(
                      BN_INIT_DECAY,
                      batch*BATCH_SIZE,
                      BN_DECAY_DECAY_STEP,
                      BN_DECAY_DECAY_RATE,
                      staircase=True)
    bn_decay = tf.minimum(BN_DECAY_CLIP, 1 - bn_momentum)
    return bn_decay
    
def get_DWA_weight(rate1, rate2, T):
    w_1 = 2*tf.math.exp(rate1 / T) / (tf.math.exp(rate1/T) + tf.math.exp(rate2/T))
    w_2 = 2*tf.math.exp(rate2 / T) / (tf.math.exp(rate1/T) + tf.math.exp(rate2/T))
    return w_1,w_2
    


def train():
    with tf.Graph().as_default():
        
        T = 1
        with tf.device('/gpu:'+str(GPU_INDEX)):
            # global LOSS_CLS_1#tf.Variable(1.0)
            # global LOSS_CLS_2  #tf.Variable(1.0)
            # global LOSS_SEG_1#tf.Variable(1.0)
            # global LOSS_SEG_2 #tf.Variable(1.0)
            pointclouds_pl, labels_pl, masks_pl, train_masks_pl, select_masks_pl = MODEL.placeholder_inputs(BATCH_SIZE, NUM_POINT)            
            is_training_pl = tf.placeholder(tf.bool, shape=())
            print(is_training_pl)
            
            LOSS_CLS_1 = tf.Variable(1.0) #tf.Variable(1.0)
            LOSS_CLS_2 = tf.Variable(1.0)  #tf.Variable(1.0)
            LOSS_SEG_1 = tf.Variable(1.0) #tf.Variable(1.0)
            LOSS_SEG_2 = tf.Variable(1.0) #tf.Variable(1.0)
        
        
            # Note the global_step=batch parameter to minimize. 
            # That tells the optimizer to helpfully increment the 'batch' parameter for you every time it trains.
            batch = tf.Variable(0)
            bn_decay = get_bn_decay(batch)
            tf.summary.scalar('bn_decay', bn_decay)
            # Get model and loss 
            class_pred, seg_pred = MODEL.get_model(pointclouds_pl, is_training_pl, bn_decay=bn_decay, num_class=NUM_CLASSES)
            total_loss, classify_loss, seg_loss,weight_p,weight_p2 = MODEL.get_loss(class_pred, seg_pred, labels_pl, masks_pl, train_masks_pl, select_masks_pl, SEG_WEIGHT,LOSS_CLS_1,LOSS_CLS_2,LOSS_SEG_1,LOSS_SEG_2)
            
            #weight_p = tf.convert_to_tensor(w1)
            #weight_p2 = tf.convert_to_tensor(w2)
            tf.summary.scalar('weight_p2',weight_p2 )
            tf.summary.scalar('weight_p',weight_p )
            
            tf.summary.scalar('total_loss', total_loss)
            tf.summary.scalar('classify_loss', classify_loss)
            tf.summary.scalar('seg_loss', seg_loss)
            
            LOSS_CLS_2 = tf.assign(LOSS_CLS_2,LOSS_CLS_1)
            LOSS_CLS_1 = tf.assign(LOSS_CLS_1,classify_loss)
            LOSS_SEG_2 = tf.assign(LOSS_SEG_2,LOSS_SEG_1)
            LOSS_SEG_1 = tf.assign(LOSS_SEG_1,seg_loss)
            
            correct = tf.equal(tf.argmax(class_pred, 1), tf.to_int64(labels_pl))
            accuracy = tf.reduce_sum(tf.cast(correct, tf.float32)) / float(BATCH_SIZE)
            tf.summary.scalar('accuracy', accuracy)

            seg_correct = tf.equal(tf.argmax(seg_pred, 2), tf.to_int64(masks_pl))
            seg_accuracy = tf.reduce_sum(tf.cast(correct, tf.float32)) / (float(BATCH_SIZE)*NUM_POINT)
            tf.summary.scalar('seg_accuracy', seg_accuracy) 

            # Get training operator
            learning_rate = get_learning_rate(batch)
            tf.summary.scalar('learning_rate', learning_rate)
            if OPTIMIZER == 'momentum':
                optimizer = tf.train.MomentumOptimizer(learning_rate, momentum=MOMENTUM)
            elif OPTIMIZER == 'adam':
                optimizer = tf.train.AdamOptimizer(learning_rate)
            train_op = optimizer.minimize(total_loss, global_step=batch)
            
            # Add ops to save and restore all the variables.
            saver = tf.train.Saver()
            
        # Create a session
        config = tf.ConfigProto()
        config.gpu_options.allow_growth = True
        config.allow_soft_placement = True
        config.log_device_placement = False
        sess = tf.Session(config=config)

        # Add summary writers
        merged = tf.summary.merge_all()
        train_writer = tf.summary.FileWriter(os.path.join(LOG_DIR, 'train'),
                                  sess.graph)
        test_writer = tf.summary.FileWriter(os.path.join(LOG_DIR, 'test'))

        # Init variables
        init = tf.global_variables_initializer()
        # To fix the bug introduced in TF 0.12.1 as in
        # http://stackoverflow.com/questions/41543774/invalidargumenterror-for-tensor-bool-tensorflow-0-12-1
        #sess.run(init)
        sess.run(init, {is_training_pl: True})

        ops = {'pointclouds_pl': pointclouds_pl,
               'labels_pl': labels_pl,
               'masks_pl': masks_pl,
               'train_masks_pl': train_masks_pl,
               'select_masks_pl': select_masks_pl,
               'is_training_pl': is_training_pl,
               'pred': class_pred,
               'seg_pred': seg_pred,
               'loss': total_loss,
               'classify_loss': classify_loss,
               'seg_loss': seg_loss,
               'train_op': train_op,
               'merged': merged,
               'step': batch,
               'weight_p':tf.convert_to_tensor(weight_p),
               'weight_p2':tf.convert_to_tensor(weight_p2)
        }

        for epoch in range(MAX_EPOCH):
            log_string('**** EPOCH %03d ****' % (epoch))
            sys.stdout.flush()
             
            train_one_epoch(sess, ops, train_writer)
            eval_one_epoch(sess, ops, test_writer)
            
            # Save the variables to disk.
            save_path = saver.save(sess, os.path.join(LOG_DIR, "model.ckpt"))
            log_string("Model saved in file: %s" % save_path)



def train_one_epoch(sess, ops, train_writer):
    """ ops: dict mapping from string to tf ops """
    is_training = True

    current_data, current_label, current_mask, current_train_mask, current_select_mask = data_utils.get_current_data_withmask_h5(
        TRAIN_DATA, TRAIN_LABELS, TRAIN_MASKS,Seg_train_mask, Seg_select_mask,NUM_POINT)

    current_label = np.squeeze(current_label)
    current_mask = np.squeeze(current_mask)
    current_train_mask  = np.squeeze(current_train_mask)
    current_select_mask = np.squeeze(current_select_mask)

    num_batches = current_data.shape[0]//BATCH_SIZE

    total_correct = 0
    total_seen = 0
    loss_sum = 0
    total_correct_seg = 0    
    classify_loss_sum = 0
    seg_loss_sum = 0  
    for batch_idx in range(num_batches):
        start_idx = batch_idx * BATCH_SIZE
        end_idx = (batch_idx+1) * BATCH_SIZE

        # Augment batched point clouds by rotation and jittering
        rotated_data = provider.rotate_point_cloud(current_data[start_idx:end_idx, :, :])
        jittered_data = provider.jitter_point_cloud(rotated_data)
        feed_dict = {ops['pointclouds_pl']: jittered_data,
                     ops['labels_pl']: current_label[start_idx:end_idx],
                     ops['masks_pl']: current_mask[start_idx:end_idx],
                     ops['train_masks_pl']: current_train_mask[start_idx:end_idx],
                     ops['select_masks_pl']: current_select_mask[start_idx:end_idx],
                     ops['is_training_pl']: is_training,}
        summary, step, _, loss_val, pred_val, seg_val, classify_loss, seg_loss,weight_w1,weight_w2 = sess.run([ops['merged'], ops['step'],
            ops['train_op'], ops['loss'], ops['pred'], ops['seg_pred'], ops['classify_loss'], ops['seg_loss'], ops['weight_p'],ops['weight_p2']], feed_dict=feed_dict)
        train_writer.add_summary(summary, step)
        print("w1 & w2:" + str(weight_w1) + " "+ str(weight_w2))
        print("current_loss" + str(classify_loss))
        print("current_seg loss" + str(seg_loss))
        pred_val = np.argmax(pred_val, 1)
        correct = np.sum(pred_val == current_label[start_idx:end_idx])

        seg_val = np.argmax(seg_val, 2)
        seg_correct = np.sum(seg_val == current_mask[start_idx:end_idx])
        total_correct_seg += seg_correct

        total_correct += correct
        total_seen += BATCH_SIZE
        loss_sum += loss_val
        classify_loss_sum += classify_loss
        seg_loss_sum += seg_loss          
    
    log_string('mean loss: %f' % (loss_sum / float(num_batches)))
    log_string('classify mean loss: %f' % (classify_loss_sum / float(num_batches)))
    log_string('seg mean loss: %f' % (seg_loss_sum / float(num_batches)))
    log_string('accuracy: %f' % (total_correct / float(total_seen)))
    log_string('seg accuracy: %f' % (total_correct_seg / (float(total_seen)*NUM_POINT))) 

        
def eval_one_epoch(sess, ops, test_writer):
    """ ops: dict mapping from string to tf ops """
    is_training = False
    total_correct = 0
    total_seen = 0
    loss_sum = 0
    classify_loss_sum = 0
    seg_loss_sum = 0
    total_correct_seg = 0

    total_seen_class = [0 for _ in range(NUM_CLASSES)]
    total_correct_class = [0 for _ in range(NUM_CLASSES)]

    current_data, current_label, current_mask = data_utils.get_current_testdata_withmask_h5(TEST_DATA, TEST_LABELS, TEST_MASKS, NUM_POINT)

    current_label = np.squeeze(current_label)
    current_mask = np.squeeze(current_mask)
    current_train_mask = np.array(current_data.shape[0]*[[True]*NUM_POINT])
    current_select_masks = np.array(current_data.shape[0]*[[True]*NUM_POINT])
    num_batches = current_data.shape[0]//BATCH_SIZE
        
    for batch_idx in range(num_batches):
        start_idx = batch_idx * BATCH_SIZE
        end_idx = (batch_idx+1) * BATCH_SIZE

        feed_dict = {ops['pointclouds_pl']: current_data[start_idx:end_idx, :, :],
                     ops['labels_pl']: current_label[start_idx:end_idx],
                     ops['masks_pl']: current_mask[start_idx:end_idx],
                     ops['train_masks_pl']: current_train_mask[start_idx:end_idx],
                     ops['select_masks_pl']: current_select_masks[start_idx:end_idx],
                     ops['is_training_pl']: is_training}
        summary, step, loss_val, pred_val, seg_val, classify_loss, seg_loss = sess.run([ops['merged'], ops['step'],
            ops['loss'], ops['pred'], ops['seg_pred'], ops['classify_loss'], ops['seg_loss']], feed_dict=feed_dict)
        test_writer.add_summary(summary, step)
        pred_val = np.argmax(pred_val, 1)
        correct = np.sum(pred_val == current_label[start_idx:end_idx])
        seg_val = np.argmax(seg_val, 2)
        seg_correct = np.sum(seg_val == current_mask[start_idx:end_idx])
        total_correct_seg += seg_correct        
        total_correct += correct
        total_seen += BATCH_SIZE
        loss_sum += (loss_val*BATCH_SIZE)
        classify_loss_sum += classify_loss*BATCH_SIZE
        seg_loss_sum += seg_loss*BATCH_SIZE         
        for i in range(start_idx, end_idx):
            l = current_label[i]
            total_seen_class[l] += 1
            total_correct_class[l] += (pred_val[i-start_idx] == l)
            
    log_string('eval mean loss: %f' % (loss_sum / float(total_seen)))
    log_string('eval accuracy: %f'% (total_correct / float(total_seen)))
    log_string('eval avg class acc: %f' % (np.mean(np.array(total_correct_class)/np.array(total_seen_class,dtype=np.float))))    
    log_string('eval seg accuracy: %f' % (total_correct_seg / (float(total_seen)*NUM_POINT)))
             

if __name__ == "__main__":
    train()
    LOG_FOUT.close()
